/*
 * Copyright (C) 2008 2009
 * Pablo Alvarado
 *
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   colorProbability.cpp

 *         Contains an example for the use of the class
 *         lti::colorProbabilityMap.

 * \author Pablo Alvarado
 *
 * \date   27.12.2008
 *
 * revisions ..: $Id: colorProbability.cpp,v 1.4 2009/09/19 00:18:41 alvarado Exp $
 */

#include <cstdlib>
#include <string>
#include <list>
#include <fstream>
#include <iostream>

#include <ltiHistogram.h>
#include <ltiColorProbabilityMap.h>
#include <ltiLispStreamHandler.h>
#include <ltiColorModelEstimation.h>
#include <ltiLoadImageList.h>
#include <ltiIOImage.h>
#include <ltiViewer2D.h>

// #define _WITH_V4L2

#ifdef _WITH_V4L2
#include "ltiV4l2.h"
#endif

/**
 * Abstraction class to wrap whether a camera data source, or a list of files.
 */
class source {
public:
  /**
   * Which input should be used
   */
  enum eInput {
    Files,
    V4L2
  };

  /**
   * Constructor
   */
  source(const eInput src,
         const std::list<std::string>& files = std::list<std::string>() );

  /**
   * Capture the next image
   *
   * @return \c true if a next image could be read or \c false if an error or
   */
  bool next(lti::image& img);

protected:
  /**
   * Get next image from file
   */
  bool nextImage(lti::image& img);

  /**
   * Get next image from camera
   */
  bool nextV4l2(lti::image& img);
  
  /**
   * Camera reading class
   */
#ifdef _WITH_V4L2
  lti::v4l2 v4l2_;
#endif

  /**
   * List of files to be read
   */
  std::list<std::string> files_;

  /**
   * Type of image input
   */
  eInput input_;
};

source::source(const eInput src,
               const std::list<std::string>& files) {
  files_ = files;
  input_ = src;
}

bool source::nextImage(lti::image& img) {
  std::string file;

  lti::ioImage loader;
  bool b=false;

  while (!b and !files_.empty()){
    file = files_.front();
    files_.pop_front();

    std::cout << "Analyzing '" << file << "'" << std::endl;

    b = loader.load(file,img);
  };

  return b;
}

bool source::nextV4l2(lti::image& img) {
#ifdef _WITH_V4L2
  static lti::v4l2 cam;

  return cam.apply(img);
#else
  std::cerr << "Support for V4l2 removed at compilation time." << std::endl;
  return false;
#endif
}


bool source::next(lti::image& img) {
  switch(input_) {
    case Files:
      return nextImage(img);
      break;
    case V4L2:
      return nextV4l2(img);
      break;
    default:
      return false;
  }
}


/**
 * Just a container for the example
 */
class example {
public:
  /**
   * Default constructor
   */
  example();

  /**
   * The execution method
   */
  int operator()(int argc,char *argv[]);

  /**
   * Read example configuration
   */
  bool read(lti::ioHandler& handler);

  /**
   * Write example configuration
   */
  bool write(lti::ioHandler& handler) const;
  
  /**
   * Train
   */
  bool train();

  /**
   * Classify
   */
  bool classify(const std::list<std::string>& imgFiles);

protected:
  /**
   * Default configuration file name
   */
  static const char *const defaultConfigFile_;

  /**
   * Configuration file name
   */
  std::string configurationFile_;

  /**
   * Name of file containing the object model.
   *
   * In training mode, the resulting models will be left in this file.  In
   * classification mode, the model is read from this file.
   */
  std::string objectModelFile_;

  /**
   * Name of file containing the non-object model
   *
   * In training mode, the resulting models will be left in this file.  In
   * classification mode, the model is read from this file.
   */
  std::string nonObjectModelFile_;

  /**
   * Train images file.
   *
   * This attribute contains the name of a file containing a list of image
   * files that should be used for training.
   */
  std::string trainImages_;

  /**
   * Mask suffix
   *
   * For each image in the trainImages_ file, a corresponding mask is searched
   * which will have this suffix appended.
   */
  std::string maskSuffix_;

  /**
   * Parameters
   * 
   * Parameters of the color probability map
   */
  lti::colorProbabilityMap::parameters parameters_;

  /**
   * Get the name of the mask file, given the image file
   */
  std::string getMaskName(const std::string& imgName) const;

private:
  /**
   * Model for the object
   */
  lti::dhistogram objectModel_;

  /**
   * Model for the non-object
   */
  lti::dhistogram nonObjectModel_;

  /**
   * Train mode
   */
  bool trainMode_;

  /**
   * Source of evaluated images
   */
  source::eInput sourceType_;

  /**
   * Print the usage of this example.
   */
  void usage(int argc,char *argv[]);

  /**
   * Parse the commands
   */
  bool parse(int argc,char *argv[],std::list<std::string>& files);
};

const char *const example::defaultConfigFile_ = "colorProbability.cfg";

example::example() {
  // initialization of the attributes
  configurationFile_  = defaultConfigFile_;
  objectModelFile_    = "objectModel.hist";
  nonObjectModelFile_ = "nonObjectModel.hist";
  trainImages_        = "images.txt";
  maskSuffix_         = "_mask";

  sourceType_         = source::Files;

  trainMode_ = false;
}

bool example::read(lti::ioHandler& handler) {
  bool b = true;

  b = lti::read(handler,"objectModelFile",objectModelFile_) && b;
  b = lti::read(handler,"nonObjectModelFile",nonObjectModelFile_) && b;
  b = lti::read(handler,"trainImages",trainImages_) && b;
  b = lti::read(handler,"maskSuffix",maskSuffix_) && b;
  b = lti::read(handler,"parameters",parameters_) && b;

  return b;
}

bool example::write(lti::ioHandler& handler) const {
  bool b = true;

  b = lti::write(handler,"objectModelFile",objectModelFile_) && b;
  b = lti::write(handler,"nonObjectModelFile",nonObjectModelFile_) && b;
  b = lti::write(handler,"trainImages",trainImages_) && b;
  b = lti::write(handler,"maskSuffix",maskSuffix_) && b;
  b = lti::write(handler,"parameters",parameters_) && b;

  return b;
}

void example::usage(int argc,char *argv[]) {
  std::cout << "\nUsage: " << argv[0] 
            << " [-t] [-c filename] [-h] [image1 [image2 ...] ]\n\n"; 
  std::cout << "  -t Train the models\n";
  std::cout << "  -c Use the given filename instead of the default '"
            << defaultConfigFile_ << "'\n";
  std::cout << "  -h Show this help\n";
  std::cout << "  -v Use the V4l2 class to get images from a camera"
            << std::endl;
}

bool example::parse(int argc,char *argv[],std::list<std::string>& files) {
  int i=1;

  files.clear();

  while (i<argc) {
    if (std::string(argv[i]) == "-c") {
      ++i;
      if (i<argc) {
        configurationFile_ = argv[i];
      } else {
        usage(argc,argv);
        return false; // report error
      }
    } else if (std::string(argv[i]) == "-t") {
      trainMode_ = true;
    } else if ( (std::string(argv[i]) == "-h") ||
                (std::string(argv[i]) == "--help") ) {
      usage(argc,argv);
      return false;
    } else if (std::string(argv[i]) == "-v") {
      sourceType_ = source::V4L2;
    } else {
      files.push_back(argv[i]);
    }
    ++i;
  }

  return true;
}

int example::operator()(int argc,char *argv[]) {
  std::list<std::string> argFiles;

  if (!parse(argc,argv,argFiles)) {
    return EXIT_FAILURE;
  }
  
  //
  // read the configuration file, and write it if something went wrong
  //
  std::ifstream in(configurationFile_.c_str());
  bool writeConfig = true;
  if (in) {
    lti::lispStreamHandler lsh(in);
    writeConfig = !read(lsh);
  }

  if (writeConfig) {
    std::ofstream out(configurationFile_.c_str());
    lti::lispStreamHandler lsh(out);
    if (!write(lsh)) {
      std::cerr << "Could not write file '" << configurationFile_ << "'" 
                << std::endl;

      return EXIT_FAILURE;
    }
    out << std::endl;
    out.close();
  }

  in.close();

  //
  // Now the real job!
  //
  if (trainMode_) {
    train();
  } else {
    classify(argFiles);
  }

  return EXIT_SUCCESS;
}

std::string example::getMaskName(const std::string& imgName) const {
  std::string file(imgName);
  std::string::size_type pos = file.rfind('.');
  if (pos != std::string::npos) {
    file.insert(pos,maskSuffix_);
  } else {
    file+=maskSuffix_;
  }
  return file;
}



bool example::train() {

  std::cout << "Training color models" << std::endl;

  lti::colorModelEstimation::parameters cmePar;
  cmePar.maskIgnore.set(32,192);
  lti::colorModelEstimation objEstimator(cmePar), nonObjEstimator(cmePar);

  lti::ioImage maskLoader;
  lti::loadImageList loader;
  loader.useFileList(trainImages_);
  lti::image img;
  lti::channel8 mask;
  int n=0; // number of read images
  while (loader.hasNext()) {
    std::string file = loader.getNextFilename();
    std::cout << "  Reading '" << file << "'" << std::flush;
    if (loader.load(img)) {
      std::cout << " done." << std::endl;
      ++n;
      // try the mask
      std::string maskName = getMaskName(file);
      bool hasMask = ( maskLoader.load(maskName,mask) && 
                      (mask.size() == img.size()) );
     
      if (hasMask) {
        std::cout << "  Valid mask found." << std::endl;
        // if a valid mask exist, then consider object and non-object pixels
        objEstimator.consider(img,mask,nonObjEstimator);
      } else {
        // if no mask available, consider all pixels as part of the object.
        objEstimator.consider(img,mask);
      }
    } else {
      std::cout << " failed: " << loader.getStatusString() << std::endl;
    }
  }

  if (n>0) {
    objEstimator.apply(objectModel_);
    nonObjEstimator.apply(nonObjectModel_);

    // now save the models
    std::ofstream outObj(objectModelFile_.c_str());
    lti::lispStreamHandler lsh(outObj);
    if (!objectModel_.write(lsh)) {
      std::cerr << "Could not write '" <<
        objectModelFile_ << "': " << lsh.getStatusString() << std::endl;
    }
    outObj << std::endl;
    outObj.close();
    
    std::ofstream outNObj(nonObjectModelFile_.c_str());
    lsh.use(outNObj);
    if (!nonObjectModel_.write(lsh)) {
      std::cerr << "Could not write '" <<
        nonObjectModelFile_ << "': " << lsh.getStatusString() << std::endl;
    }
    outNObj << std::endl;
    outNObj.close();
    
    return true;
  } else {
    std::cerr << "No image loaded.  Train failed!" << std::endl;
  }
  
  return false;
}


bool example::classify(const std::list<std::string>& imgFiles) {
  // first, read both models
  std::ifstream inObj(objectModelFile_.c_str());
  lti::lispStreamHandler lsh(inObj);
  if (!objectModel_.read(lsh)) {
    std::cerr << "Could not read object model file '" << objectModelFile_
              << "'" << std::endl;
  }

  inObj.close();
    
  std::ifstream inNObj(nonObjectModelFile_.c_str());
  lsh.use(inNObj);
  if (!nonObjectModel_.read(lsh)) {
    std::cerr << "Could not read non-object model file '" << nonObjectModelFile_
              << "'" << std::endl;
  }
  inNObj.close();





  // second, initialize the color probability functor

  lti::colorProbabilityMap mapper(parameters_);
  mapper.setColorModels(objectModel_,nonObjectModel_);

  // now, just generate the maps
  // can be
  // - an image
  // - a camera stream
  source src(sourceType_,imgFiles);

  lti::image img; 
  lti::channel cmap;
  lti::viewer2D viewOrg("Input"),viewMap("Probability");
  lti::viewer2D::interaction action;
  lti::ipoint position;

  // as long as there are available images
  while (src.next(img)) {
    mapper.apply(img,cmap);
    viewOrg.show(img);
    viewMap.show(cmap);
    if (sourceType_ == source::Files) {
      do {
        viewMap.waitInteraction(action,position);
      } while(action.action != lti::viewer2D::Closed);
    }
  }

  viewOrg.hide();
  viewMap.hide();

  return true;
}


int main(int argc,char *argv[]) {
  try {
    example obj;
    return obj(argc,argv);
  }
  catch (lti::exception& exp) {
    std::cout << "An LTI::EXCEPTION was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (std::exception& exp) {
    std::cout << "std::exception was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (...) {
    std::cout << "Unknown exception thrown!" << std::endl;
  }
  return EXIT_FAILURE;

} // main
